# -*- coding: utf-8 -*-

# T. Ida, Sep. 8, 2021

import numpy as np # import "numpy" module
from scipy import interpolate # import "interploate" module from "scipi"
import os.path # import "os.path" module
import configparser # import "configparser" module

import dct_common as cmn # import "dct_common.py" as "cmn"

#####################################################
# calc_kappas()
# Calculate Cumulants of Axial-Divergence Aberration
# Output units are [1,rad.,rad.^2,rad.^3,rad.^4]
#####################################################
def calc_kappas(degTwot, degAxial):
    # degTwot: 2Theta in [deg.] (NumPy array)
    # degAxial: Axial divergence angle (float number)
    theta = degTwot * np.pi / 360    # (NumPy array)
    tanT = np.tan(theta) # (NumPy array)
    cotT = 1.0 / tanT # (NumPy array)
    psi2 = (degAxial * np.pi / 180)**2 # (float)
    nSize = degTwot.size
    s0 = np.ones(nSize)
    # Calculate Cumulants up to 4th order
    kappa1 = (psi2/12)*(tanT - cotT)
    kappa2 = tanT**2 + 6.0/17 + cotT**2
    kappa2 *= 17 * psi2**2 / 1440
    kappa3 = tanT**3-cotT**3 + 81*(tanT-cotT)/169
    kappa3 *= 169 * psi2**3 / 60480
    kappa4 = tanT**4+cotT**4 + 0.560143*(tanT**2+cotT**2) - 0.462679
    kappa4 *= 0.000801577 * psi2**4
    # Convert to Reduced Cumulants in [deg.]
    kappa1 = 180.0/np.pi*kappa1
    kappa2 = 180.0/np.pi*np.sqrt(kappa2)
    kappa3 = 180.0/np.pi*np.sign(kappa3)*np.abs(kappa3)**(1/3)
    kappa4 = 180.0/np.pi*np.sign(kappa4)*np.abs(kappa4)**(1/4)
    kappas = np.array([s0, kappa1,kappa2,kappa3,kappa4])
    return([s0,kappa1,kappa2,kappa3,kappa4])

#############################################
# Define Horizontal Scale Transform Function
#############################################
def scale_hrz(sign, degTwoTheta):
    # sign = 1: higher-angle component
    # sign = -1: lower-angle component
    bet = 0.197562 # = (71 - 14*np.sqrt(22))/27
    x = sign * np.log(1 + bet - sign * (1 - bet) * np.cos(np.pi * degTwoTheta / 180))
    x = x / (1 - bet)
    return x

#########################################
# Define Ordinate Scale Transform Factor
#########################################
def scale_ord(sign, degTwoTheta):
    # sign = 1: higher-angle component
    # sign = -1: lower-angle component
    bet = 0.197562 # = (71 - 14*np.sqrt(22))/27
    t = np.tan(np.pi * degTwoTheta / 360)
    if (sign == -1):
        f = t + bet / t
    else:
        f = bet * t + 1 / t
    return f


###################################################################
# Define Fourier Transform of Peak Profile Model for Deconvolution
# for sign == 1/-1, 
###################################################################
def fFtDeconv(radPsi, sign, xi):
    # Psi = 0.04 # rad
    # sign = 1, -1
    sqPsi = radPsi * radPsi
    beta = 0.197562 # = (71 - 14*np.sqrt(22))/27
    alpha = 0.80
    gamma = np.power(169/120960/alpha/(1-np.power(beta,3)),1.0/3) * sqPsi
    delta = sqPsi / (12*(1-beta)) - alpha * gamma
    ans = np.exp(-sign * 2*np.pi*1j*xi*delta) 
    ans = ans * np.power(1 + sign * 2*np.pi*1j*xi*gamma,-alpha)
    return ans


######################################################
# "axial.treat"
# TREATMENT OF AXIAL DIVERGENCE ABERRATION, MAIN BODY
######################################################
def treat(sign, source, DPATH):
    # sign = 1: higher-angle / sign = -1: lower-angle
    # source: Source Data (Python List)
    #    [s_angle, s_intensity],
    #    [s_angle, s_intensity, s_error1], or
    #    [s_angle, s_intensity, s_error1, s_error2]
    # DPATH: Path of Configuration File
    
    #***********************************
    # Read Configuration File "dct.cfg"
    #***********************************
    CFG_FILE = os.path.join(DPATH, 'dct.cfg')
    dct_config = configparser.ConfigParser()
    dct_config.read(CFG_FILE, encoding = 'utf-8')
    error_type = dct_config.get('Control','error_type')
    # Axial-divergence angle (deg.)
    degPsi = dct_config.getfloat('Control','degAxial')
    # Axial-divergence angle (rad)
    radPsi = degPsi * np.pi / 180
    # Margin for Fourier-based Deconvolutional Treatment
    marginAxial = dct_config.getfloat('Control','marginAxial')

    #*****************************************************
    # Assign Items in Python List "source" to numpy.array
    #*****************************************************
    s_angle = np.array(source[0]) # 0th item is array of angles
    s_intensity = np.array(source[1]) # 1st item is array of intensity
    d_intensity = np.array(s_intensity)

    #----------------------------------------
    # Optional Output for 'skipAxial' == '2'
    #----------------------------------------
    CFG_FILE = 'dct.cfg'
    dct_config = configparser.ConfigParser()
    dct_config.read(CFG_FILE, encoding = 'utf-8')
    skipAxial = '0'
    if (dct_config.has_option('Control','skipAxial')):
        skipAxial = dct_config.get('Control','skipAxial')
    if (skipAxial == '2'):
        kappas = calc_kappas(s_angle, degPsi)
        CUMULANT_FILE = 'cumulants_axial.csv'
        save_cumulant = np.insert(kappas, 0, s_angle, axis=0)
        save_cumulant = np.array(save_cumulant).T    # transpose matrix
        fmt_list = ["%.3f","%.5e","%.5e","%.5e","%.5e","%.5e"]
        np.savetxt(CUMULANT_FILE, save_cumulant, delimiter=",",fmt=fmt_list)

    #*****************************************************
    # Step 1: Scale Transform, "chi"
    # s_angle (2Theta in deg.) => source_x (on chi-scale)
    # s_intensity => source_y (on chi3-scale)
    #*****************************************************
    source_x = scale_hrz(sign, s_angle) # horizontal
    source_corr = scale_ord(sign, s_angle)
    source_corr = source_corr * cmn.g_corr(s_angle / 2) # correction factor
    source_y = s_intensity * source_corr # ordinate

    #******************************************************************
    # Step 2: Preparation for Fourier treatment
    # Determine Marginal (Padding) Range, Total Points &amp; Step Interval
    #******************************************************************
    margin = marginAxial
    index = cmn.indices(source_x, margin)
    nSource = index[0]
    i1 = index[1] # 1st index
    i2 = index[2] # 2nd index
    i3 = index[3] # 3rd index
    nDest = index[4]
    nDest_2 = int(nDest / 2)
    # Determine Step Interval, dx
    dx = (source_x[nSource-1] - source_x[0]) / i1
    # Prepare Arrays for Interpolated Data in Valid Range
    nValid = int((source_x[nSource-1] - source_x[0]) / dx)
    chi0 = source_x[0]
    chi = np.linspace(chi0, chi0 + nValid * dx, nValid, endpoint = False)

    #*********************************************************
    # Step 3: 
    # Cubic spline interpolation of Data to Transformed Scale
    # (source_x, source_y) => (chi, eta)
    #*********************************************************
    # Cubic spline interpolation
    f3 = interpolate.interp1d(source_x, source_y, kind="cubic")
    eta = f3(chi)

    #********************************************
    # Step 4: 
    # Extrapolation for marginal (padding) range
    # (chi, eta) => (chi, eta)
    #********************************************
    # Fill Marginal Area with Values
    chi = np.linspace(source_x[0], source_x[0] + nDest * dx, nDest, endpoint = False)
    eta = np.resize(eta, nDest)
    iLin = np.linspace(0, i3-i2, i3-i2+1)
    etaL = source_y[nSource - 1]
    etaR = source_y[0]
    eta[i1:i2] = etaL
    eta[i2:i3] = etaL + (etaR - etaL) * iLin[0:i3-i2] / (i3-i2)
    eta[i3:nDest] = etaR

    #*********************************************
    # Step 5: Fourier Transform of Intensity Data
    # (eta) => (ft_eta)
    #*********************************************
    # Fourier transform of intensity data
    ft_eta = np.fft.fft(eta)
    # Step interval of Fourier transform
    dxi = 1.0 / (dx * nDest)
    # Prepare Abscissa Data Using compliment expression
    # Abscissa of Fourier transform
    xi = np.linspace(0, nDest * dxi, nDest, endpoint=False)
    # Use compliment expression for Fourier transform
    xi[nDest_2:nDest] = xi[nDest_2:nDest] - nDest * dxi

    #**********************************************************
    # Step 6: Fourier Transform of Approximate Model Profile
    # Function To Be Deconvolved (+/- Axial Model)
    # => (ft_deconv), (ft_conv)
    #**********************************************************
    # Fourier Transform of Function To Be Deconvolved (+/- Axial)
    ft_deconv = fFtDeconv(radPsi, sign, xi)
    # Fourier transform of instrumental function
    ft_conv = np.abs(ft_deconv) # (absolute value)
    # Fourier transform of instrumental function
    ft_inst = ft_deconv / ft_conv

    #**********************************************************
    # Step 7: Here is the Core of DCT,
    # Deconvolution/Convolution Treatment on Fourier transform
    # (ft_eta)/(ft_inst) => (ft_eta)
    #**********************************************************
    # Deconvolution/Convolution in Fourier transform
    ft_eta = ft_eta / ft_inst

    #***********************************
    # Step 8: Inverse Fourier Transform
    # (ft_eta) => (zeta)
    #***********************************
    # Inverse Fourier transform of DCT data
    zeta = np.real(np.fft.ifft(ft_eta))
    
    #***********************************************
    # Step 9: Pick up Intensity from DCT data
    # on "chi" scale by cubic spline interpolation
    # (chi, zeta) => (source_x, dest_y)
    #***********************************************
    # Reset equidistant abscissa
    chi = np.linspace(chi0, chi0 + nDest * dx, nDest, endpoint = False)
    # Cubic spline interpolation, (chi,zeta) => (source_x, dest_y)
    f3 = interpolate.interp1d(chi, zeta, kind="cubic")
    dest_y = f3(source_x)

    #**********************************************
    # Step 10: Rescale transform
    # (source_x, dest_y) => (s_angle, d_intensity)
    #**********************************************
    # Rescale transform
    d_intensity = dest_y / source_corr # ordinate values
    dest = source
    dest[1] = d_intensity
    
    #***********************************************************
    # Step 11: Fourier Treatment of Error Data on axial.treat()
    #***********************************************************
    dest = cmn.TreatErrors(error_type, source, source_x, index, source_corr, ft_inst)

    return dest
    
