# -*- coding: utf-8 -*-
import numpy as np # import "numpy" module
from scipy import interpolate # import "interpolate" from "scipy"
import datetime # import "datetime" module

# Python codes commonly used for deconvolutional treatment

#***********************************************
# Find Total Points and Indices of Anker Points
# from Non-equidistant Data &amp; Margin Width
#***********************************************
def indices(x, w_margin):
    # x[..] # abscissa values of non-equidistant data (numpy array)
    # w_margin = 0.5 # margin width
    nSource = len(x)
    dx = np.zeros(nSource-1)
    dx = x[1:nSource-1] - x[0:nSource-2]
    iMin = np.ndarray.argmin(dx)
    dx_min = (x[iMin+1] - x[iMin]) / 2
    w_valid = x[nSource-1] - x[0]
    power2 = np.log2((w_valid + w_margin) / dx_min)
    power2 = int(np.ceil(power2)) # ceiling function
    nDest = np.power(2, power2) # total number
    dx = (w_valid + w_margin) / nDest
    i1 = int(np.round(w_valid / dx))
    i2 = int(np.round((2 * i1 + nDest) / 3))
    i3 = int(np.round((i1 + 2 * nDest) / 3))
    return [nSource, i1, i2, i3, nDest]

###################################################
# Padding for Marginal Range for Fourier Treatment 
###################################################
def pad_margin(y, index):
    # y: y-array (NumPy array)
    # index: 5-item list (Python List)
    i1 = index[1]    # end of valid range
    i2 = index[2]    # 1st anker
    i3 = index[3]    # 2nd anker
    i4 = index[4]    # total number
    iLin = np.linspace(0, i3-i2, i3-i2)
    yL = y[i1-1]
    yR = y[0]
    y[i1:i2] = yL
    y[i2:i3] = yL + (yR - yL)*iLin[0:i3-i2]/(i3-i2)
    y[i3:i4] = yR

#################################################
# Modification of Errors for Change in Intervals
#################################################
def mod_error(x, d, dx):
    n = x.size
    e = np.array(d) # make a copy
    e[0] = d[0] * np.sqrt((x[1]-x[0]) / dx)
    for i in range(1, n - 1):
        e[i] = d[i] * np.sqrt((x[i+1]-x[i-1]) / (2*dx))
    e[n-1] = d[n-1] * np.sqrt((x[n-1]-x[n-2]) / dx)
    return e

####################################################
# Re-Modification of Errors for Change in Intervals
####################################################
def remod_error(x, e, dx):
    n = x.size
    d = np.array(e) # make a copy
    d[0] = e[0] / np.sqrt((x[1]-x[0]) / dx)
    for i in range(1, n - 1):
        d[i] = e[i] / np.sqrt((x[i+1]-x[i-1]) / (2*dx))
    d[n-1] = e[n-1] / np.sqrt((x[n-1]-x[n-2]) / dx)
    return d

###########################################################
# Geometrical Correction Factor for
# Bragg-Brentano Diffractometer Without Diffracting Optics
###########################################################
def g_corr(degTheta):
    radTheta = degTheta * np.pi / 180
    ans = 2 * np.sin(radTheta) * np.sin(2 * radTheta)
    ans = ans / (1 + np.square(np.cos(2 * radTheta)))
    return ans

##############################
# Fourier Treatment of Errors
##############################
def TreatErrors(error_type, source, source_x, index, source_corr, ft_inst):
        # error_type: type of error models [0,1,2,3]
        # source: source data (Python list of arrays)
        # source_x: scale-transformed abscissa
        # index: Python list of anker points
        # source_corr: correction factor
        # ft_inst: Fourier transform of instrumental function
    dest = source
    nSource = index[0]
    i1 = index[1] # 1st index
    i2 = index[2] # 2nd index
    i3 = index[3] # 3rd index
    nDest = index[4]
    nDest_2 = int(nDest / 2)
    # Determin Step Interval, dx
    dx = (source_x[nSource-1] - source_x[0]) / i1
    # Prepare Arrays for Interpolated Data in Valid Range
    nValid = int((source_x[nSource-1] - source_x[0]) / dx)
    chi0 = source_x[0]
    # Linear interpolation function
    iLin = np.linspace(0, i3-i2, i3-i2+1)
    #********************
    # Pick up error data
    #********************
    if (error_type == '1'):
        s_error1 = np.array(source[2])
    elif (error_type == '2'):
        s_error2 = np.array(source[2])
    elif (error_type == '3'):
        s_error1 = np.array(source[2])
        s_error2 = np.array(source[3])
    #******************************************
    # Fourier Treatment of Error Data (type I)
    #******************************************
    if ((error_type == '1') or (error_type == '3')):
        # Correct errors for scale transform &amp; geometrical factors
        source_d1 = s_error1 * source_corr
        # Correction for change in interval
        source_e = cmn.mod_error(source_x, source_d1, dx)
        # Abscissa points for interpolation
        chi = np.linspace(chi0, chi0 + nValid * dx, nValid, endpoint = False)
        # Cubic spline interpolation
        f3 = interpolate.interp1d(source_x, source_e, kind="cubic")
        eps = f3(chi)
        epsL = source_e[nSource - 1]
        epsR = source_e[0]
        # Add marginal range
        eps = np.resize(eps, nDest)
        # Fill marginal range with values
        eps[i1:i2] = epsL
        eps[i2:i3] = epsL + (epsR - epsL) * iLin[0:i3-i2] / (i3-i2)
        eps[i3:nDest] = epsR
        # Inverse squared errors
        inv_sq_eps = 1.0 / (eps * eps)
        # Fourier transform of inverse squared errors
        ft_inv_sq_eps = np.fft.fft(inv_sq_eps)
        # Instrumental function
        inst = np.real(np.fft.ifft(ft_inst))
        # Squared instrumental function
        sq_inst = inst * inst
        # Fourier transform of squared instrumental function
        ft_sq_inst = np.fft.fft(sq_inst)
        # Mutual correlation in Fourier transform
        ft_inv_sq_eps = ft_inv_sq_eps * np.conj(ft_sq_inst)
        # Inverse Fourier transform
        delta = np.real(np.fft.ifft(ft_inv_sq_eps))
        # Estimated error (model I)
        delta = 1 / np.sqrt(np.abs(delta))
        # Abscissa of Fourier transform
        chi = np.linspace(chi0, chi0 + nDest * dx, nDest, endpoint = False)
        # Cubic spline interpolation
        f3 = interpolate.interp1d(chi, delta, kind="cubic")
        dest_e = f3(source_x)
        # Recorrect error values for change in intervals
        dest_d = cmn.remod_error(source_x, dest_e, dx)
        # Recorrect error data for scale transform &amp; geometrical factors
        d_error1 = dest_d / source_corr

    #*******************************************
    # Fourier Treatment of Error Data (type II)
    #*******************************************
    if ((error_type == '2') or (error_type == '3')):
        # Correct error data for scale transform &amp; geometrical factors
        source_d2 = s_error2 * source_corr
        # Correct error values for change in intervals
        source_e = cmn.mod_error(source_x, source_d2, dx)
        # Abscissa of Fourier transform
        chi = np.linspace(chi0, chi0 + nValid * dx, nValid, endpoint = False)
        # Cubic spline interpolation
        f3 = interpolate.interp1d(source_x, source_e, kind="cubic")
        eps = f3(chi)
        epsL = source_e[nSource - 1]
        epsR = source_e[0]
        # Add marginal range
        eps = np.resize(eps, nDest)
        # Fill marginal range with values
        eps[i1:i2] = epsL
        eps[i2:i3] = epsL + (epsR - epsL) * iLin[0:i3-i2] / (i3-i2)
        eps[i3:nDest] = epsR
        # Inverse instrumental function
        inv_inst = np.real(np.fft.ifft(1 / ft_inst))
        # Squared inverse instrumental function
        sq_inv_inst = inv_inst * inv_inst
        # Fourier transform of squared inversse instrumental function
        ft_sq_inv_inst = np.fft.fft(sq_inv_inst)
        # Squared errors
        sq_eps = eps * eps
        # Fourier transform of squared errors
        ft_sq_eps = np.fft.fft(sq_eps)
        # Convolution in Fourier transform
        ft_sq_eps = ft_sq_eps * ft_sq_inv_inst
        # Invers Fourier transform
        sq_eps = np.real(np.fft.ifft(ft_sq_eps))
        # Estimated error in interpolated data
        delta = np.sqrt(np.abs(sq_eps))
        # Abscissa of Fourier transform
        chi = np.linspace(chi0, chi0 + nDest * dx, nDest, endpoint = False)
        # Cubic spline interpolation
        f3 = interpolate.interp1d(chi, delta, kind="cubic")
        dest_e = f3(source_x)
        # Recorrect error values for change in intervals
        dest_d = cmn.remod_error(source_x, dest_e, dx)
        # Recorrect error data for scale transform &amp; geometrical factors
        d_error2 = dest_d / source_corr
    #********************
    # Replace error data
    #********************
    if (error_type == '1'):
        dest[2] = d_error1
    elif (error_type == '2'):
        dest[2] = d_error2
    elif (error_type == '3'):
        dest[2] = d_error1
        dest[3] = d_error2
    return dest

##############
# Report Time
##############
def ReportTime(s):
    # Argument s: strings to be shown after date time
    d = datetime.datetime.today() # current date time
    # Show current date time
    print('{0:02d}:{1:02d}:{2:02d}'.format(d.hour, d.minute, d.second),' ',s)
    return 0
