# -*- coding: utf-8 -*-

# Coded by T. Ida, corrected by N. Sakurai
# Updated by T. Ida, Oct. 23, 2021

import numpy as np # import "numpy" module
from scipy import interpolate # import "interpolate" from "scipy"
from scipy import special # import "special" from "scipy"
import os.path # import "os.path" module
import configparser # import "configparser"
import dct_common as cmn # import "dct_common.py" as "cmn"

#################################################
# calc_kappas()
# Calculate Cumulants of Equatorial Aberration
# Output units are [1,rad.,rad.^2,rad.^3,rad.^4]
#################################################
def calc_kappas(degTwot, degDS, degPsi, W_2R):
    # degTwot: 2Theta in [deg.] (NumPy array)
    # degDS: Divergence slit angle in [deg.] (float number)
    # degPsi: Half of the SSXD view angle in [deg.] (float number)
    # W_2R: ratio of half of specimen width to goniometer radius (float number)

    #*******************************************
    # Exact Expression of Deviation Angle
    # Caused by Equatorial Aberration for given 
    # values of (theta, phi, psi)
    # Output is in [rad.]
    #*******************************************
    def deltaTwot(theta, phi, psi):
        # theta: Half of 2Theta in [rad.] (NumPy array)
        # phi: equatorial deviation angle of incident beam in [rad.] (NumPy array)
        # psi: half of offset angle of detector strip in [rad.] (float number)
        g = np.cos(theta-psi) - np.sin(theta-psi) / np.tan(theta-psi-phi)
        g = g * np.cos(2 * psi)
        g = g + np.cos(theta + psi)
        deltaTwot = theta + phi + psi
        deltaTwot -= np.arctan(np.sin(theta + psi) / g)
        return deltaTwot

    nSize = degTwot.size    # number of values of degTwot
    theta = degTwot * np.pi / 360    # shape of theta is (nSize) here
    sinTheta = np.sin(theta)    # shape of sinTheta is (nSize) here
    phi = degDS * np.pi / 180    # float number
    psi = degPsi * np.pi / 180    # float number
        # Note: effective equatorial divergence angle should be
        # reduced from divergence slit open angle Phi_DS
        # at 2Theta angles lower than 2ThetaC.
    theta = theta.repeat(4).reshape(nSize,4) # shape of theta is (nSize,4)
    # 4-point Gauss-Legendre abscissa & weights
    x_j = np.array([ 0.069431844203,0.330009478208,0.669990521792,0.930568155797 ])
    w_j = np.array([ 0.173927422569,0.326072577431,0.326072577431,0.173927422569 ])
    # Set values of psi
    psi_j = -psi/2 + x_j * psi    # shape of psi_j is (4)
    psi_j = np.tile(psi_j, nSize).reshape(nSize,4) # shape of psi_j is (nSize,4)
    # Set limit values of phi for each of (theta, psi)
    phiMin = theta - psi_j - np.arctan(np.sin(theta-psi_j)/(np.cos(theta-psi_j)-W_2R))
    phiMin = np.maximum(-phi/2, phiMin)    # shape of phiMin is (nSize, 4)
    phiMax = theta - psi_j - np.arctan(np.sin(theta-psi_j)/(np.cos(theta-psi_j)+W_2R))
    phiMax = np.minimum(phi/2, phiMax)    # shape of phiMax is (nSize, 4)
    # Set matrix elements ...
    phi0 = np.tile(phiMin,4).reshape(nSize,4,4)
    phi1 = np.tile(phiMax-phiMin,4).reshape(nSize,4,4)
    phix = np.tile(x_j.repeat(4),nSize).reshape(nSize,4,4)
    phi_ij = phi0 + phix * phi1
    theta = np.tile(theta, 4).reshape(nSize,4,4)    # shape of theta is (nSize,4,4)
    psi_ji = np.tile(psi_j, 4).reshape(nSize,4,4)    # shape of psi_j is (nSize,4,4)
    # For debug ...

    deltaTwot_ij = deltaTwot(theta, phi_ij, psi_ji)
    # print("deltaTwot_ij = ",deltaTwot_ij)
    # results should be ...
    # / d2theta(phi1,psi1) d2theta(phi1,psi2) d2theta(phi1,psi3) d2theta(phi1,psi4) \
    # | d2theta(phi2,psi1) d2theta(phi2,psi2) d2theta(phi2,psi3) d2theta(phi2,psi4) |
    # | d2theta(phi3,psi1) d2theta(phi3,psi2) d2theta(phi3,psi3) d2theta(phi3,psi4) |
    # \ d2theta(phi4,psi1) d2theta(phi4,psi2) d2theta(phi4,psi3) d2theta(phi4,psi4) /
    # ...
    deltaTwot0_ij = np.array([1.0]).repeat(nSize*4*4).reshape(nSize,4,4)
    deltaTwot2_ij = deltaTwot_ij**2    # not matrix, but element operation
    deltaTwot3_ij = deltaTwot2_ij * deltaTwot_ij    # element operation
    deltaTwot4_ij = deltaTwot2_ij**2    # element operation
    s0 = np.dot(np.dot(phi1 * deltaTwot0_ij, w_j.T) / phi, w_j.T)
    s1 = np.dot(np.dot(phi1 * deltaTwot_ij, w_j.T) / phi, w_j.T)
    s2 = np.dot(np.dot(phi1 * deltaTwot2_ij, w_j.T) / phi, w_j.T)
    s3 = np.dot(np.dot(phi1 * deltaTwot3_ij, w_j.T) / phi, w_j.T)
    s4 = np.dot(np.dot(phi1 * deltaTwot4_ij, w_j.T) / phi, w_j.T)

    # Calculate Cumulants up to 4th order
    kappa1 = s1/s0
    kappa2 = s2/s0 - (s1/s0)**2
    kappa3 = s3/s0 - 3*s2*s1/s0**2 + 2*(s1/s0)**3
    kappa4 = s4/s0 - 4*s3*s1/s0**2 - 3*(s2/s0)**2 + 12*(s2/s0)*(s1/s0)**2 - 6*(s1/s0)**4
    # Convert to Reduced Cumulants in [deg.]
    kappa1 = 180.0/np.pi*kappa1
    kappa2 = 180.0/np.pi*np.sqrt(kappa2)
    kappa3 = 180.0/np.pi*np.sign(kappa3)*np.abs(kappa3)**(1/3)
    kappa4 = 180.0/np.pi*np.sign(kappa4)*np.abs(kappa4)**(1/4)
    kappas = np.array([s0, kappa1,kappa2,kappa3,kappa4])
    return kappas

##############################################################
# Define Equatorial Aberration Model Function for "chi" scale
# (Approximate)
##############################################################
def FormFunc3(chi, rho):
    # chi: NumPy array
    # rho: float number
    chiMin = -3 * ( 1 + rho )
    if ( rho < 2 ):
        chiMax = 0.75 * rho ** 2
    else:
        chiMax = -3 * ( 1 - rho )
    if ( rho == 0 ):
        y = np.where((-3 < chi)&(chi < 0), 1/(6*np.sqrt(-chi/3)), 0) 
    else:
        D = rho**2 - chi / 0.75
        sqrtD = np.sqrt(np.maximum(0, D))
        phiL = np.maximum(-rho + sqrtD, rho - sqrtD )
        phiU = np.minimum(2, rho + sqrtD )
        y = np.where((D>0)&(chiMin<chi)&(chi<chiMax), np.log(phiU/phiL)/(6*rho), 0)
    return y

################################################
# "equatorial.treat"
# TREATMENT OF EQUATORIAL ABERRATION, MAIN BODY
################################################
def treat(source, DPATH):
    # source: Source Data (Python List)
    #    [s_angle, s_intensity],
    #    [s_angle, s_intensity, s_error1], or
    #    [s_angle, s_intensity, s_error1, s_error2]
    # DPATH: Path of Configuration File

    #***********************************
    # Read Configuration File "dct.cfg"
    #***********************************
    CFG_FILE = os.path.join(DPATH, 'dct.cfg')
    dct_config = configparser.ConfigParser()
    dct_config.read(CFG_FILE, encoding = 'utf-8')
    # Indication the way to treat error
    error_type = dct_config.get('Control','error_type')
    # Divergence Slit Open Angle (deg.), degDS (1.25, for example)
    degDS = dct_config.getfloat('Control','degDS')
    # LPSD View Angle (deg.), degLPSD (4.89, for example)
    degLPSD = dct_config.getfloat('Control','degLPSD')
    # Ratio of Specimen-Width to Goniometer Radius, specW_R
    W_R = dct_config.getfloat('Control','specW_R')
    # Margin for Fourier-based Deconvolutional Treatment
    marginEquat = dct_config.getfloat('Control','marginEquat')

    ###################################################################
    # Stage 1: Correction of the 3rd cumulant of equatorial aberration 
    # (Correction about Asymmetric Deformation)
    ###################################################################

    #******************************************************
    # Step 1-0: Divide Python List "source" to numpy.array
    #******************************************************
    s_angle = np.array(source[0])    # angle data
    s_intensity = np.array(source[1])    # intensity data
    if (error_type == '1'):
        s_error1 = np.array(source[2])
    elif (error_type == '2'):
        s_error2 = np.array(source[2])
    elif (error_type == '3'):
        s_error1 = np.array(source[2])
        s_error2 = np.array(source[3])
    d_intensity = np.array(s_intensity)

    #*******************************************************
    # Step 1-1: Scale Transform for the 3rd cumulant, "chi3"
    # s_angle (2Theta in deg.) => source_x (on chi3-scale)
    # s_intensity => source_y (on chi3-scale)
    #*******************************************************
    kappas = calc_kappas(s_angle, degDS, degLPSD/2, W_R/2)
            # Note: "degPsi" is half of the view angle degLPSD

    #---------------------------------------------
    # Optional Output for 'skipEquatorial' == '2'
    #---------------------------------------------
    CFG_FILE = 'dct.cfg'
    dct_config = configparser.ConfigParser()
    dct_config.read(CFG_FILE, encoding = 'utf-8')
    skipEquatorial = '0'
    if (dct_config.has_option('Control','skipEquatorial')):
        skipEquatorial = dct_config.get('Control','skipEquatorial')
    if (skipEquatorial == '2'):
        CUMULANT_FILE = 'cumulants_equatorial.csv'
        save_cumulant = np.insert(kappas, 0, s_angle, axis=0)
        save_cumulant = np.array(save_cumulant).T    # transpose matrix
        fmt_list = ["%.3f","%.5e","%.5e","%.5e","%.5e","%.5e"]
        np.savetxt(CUMULANT_FILE, save_cumulant, delimiter=",",fmt=fmt_list)
    kappa1 = kappas[1] * np.pi / 180
    kappa3 = np.sign(kappas[3])*np.abs(kappas[3])**3
    # print("kappa3 =",kappa3)
    rho = (degLPSD / 2) / degDS    # Note: "psi" is half of the view angle 
    k3 = -16/35*(1+21*rho**2/4)    # 3rd cumulant of model profile on chi3-scale
    dchi3_d2Theta = np.power(k3/kappa3, 1/3)    # = d(chi3)/d(2Theta)
    # Integration & creation of abscissa values source_x
    # Note: chi3 should be the integral of d(chi3)/d(2Theta) by 2Theta
    source_x = np.cumsum( dchi3_d2Theta )
    source_x *= (s_angle[1]-s_angle[0])
    # Ordinate Correction Factor
    # Note: "source_y d(chi3)" should be proportional
    #  to "s_intensity d(2Theta)"
    source_corr = dchi3_d2Theta
    # Conbine with Geometrical Correction
    # (Lorentz, polarization & powder diffractometry corrections)
    source_corr *= cmn.g_corr(s_angle / 2)
    # Another correction for loss of intensity,
    # caused by spill-out of irradiated area from specimen face
    # sinTc = np.sin(degTwotC * np.pi / 360)
    # sinT = np.sin(s_angle * np.pi / 360)
    source_corr2 = 1 / kappas[0]    #np.where(s_angle < degTwotC, sinTc/sinT, 1)
    # Corrected & Transdformed Intensity
    source_y = s_intensity * source_corr * source_corr2

    source3_x = source_x
    source3_y = source_y
    
    #******************************************************************
    # Step 1-2: Preparation for Fourier treatment
    # Determine Marginal (Padding) Range, Total Points & Step Interval
    #******************************************************************
    # marginModel: Standard deviation of model profile
    marginModel = np.sqrt(0.8*(1+1.25*rho**2))
    # margin: defines marginal (padding) range
    margin = marginEquat * marginModel
        # marginEquat is the value read from the file'dct.cfg'
        # User can change the value in the current version.
    index = cmn.indices(source_x, margin)
        # Derive total number and 1st,2nd,3rd anker points
        # dct_cmn.indices(x, w_margin) is defined in "dct_cmn.py"
        # x[..] = source_x :  abscissa (NumPy array)
    nSource = index[0]    # total number of valid data
    nValid = index[1] # number of valid data equidistantly separated
    i1 = index[2] # index of 1st anker point
    i2 = index[3] # index of 2nd anker point
    nDest = index[4]    # total number of interpolated data
    nDest_2 = int(nDest/2)    # half of the total number
    # dchi3: Step interval of interpolated data
    dchi3 = (source_x[nSource-1] - source_x[0]) / (nValid - 1)
    # Prepare equidistant data for valid range
    # nValid = int((source_x[nSource-1] - source_x[0]) / dchi3) + 1
    chi0 = source_x[0]
    chi3 = np.linspace(chi0, chi0 + nValid * dchi3, nValid, endpoint = False)

    #*********************************************************
    # Step 1-3: 
    # Cubic spline interpolation of Data to Transformed Scale
    # (source_x, source_y) => (chi3, eta3)
    #*********************************************************
    # Cubic spline interpolation
    f3 = interpolate.interp1d(source_x, source_y, kind="cubic")
    eta3 = f3(chi3)

    #********************************************
    # Step 1-4: 
    # Extrapolation for marginal (padding) range
    # (chi3, eta3) => (chi3, eta3)
    #********************************************
    # Fill marginal range with values
    chi3 = np.linspace(chi0, chi0 + nDest * dchi3, nDest, endpoint = False)
    eta3 = np.resize(eta3, nDest)
    cmn.pad_margin(eta3, index)    # => dct_common.pad_margin()

    #***********************************************
    # Step 1-5: Fourier Transform of Intensity Data
    # (eta3) => (ft_eta3)
    #***********************************************
    # Fourier transform of intensity
    ft_eta3 = np.fft.rfft(eta3)
    
    #**********************************************************
    # Step 1-6: Fourier Transform of Approximate Model Profile
    # Function To Be Deconvolved (Equatorial Model)
    # => (ft_deconv3), (ft_conv3)
    #**********************************************************
    # Prepare Fourier transform of function to be deconvolved
    # Set start point to be zero, and use compliment expression ・・
    hx = nDest * dchi3
    chi3 = np.linspace(0, hx, nDest, endpoint=False)
    chi3[nDest_2:nDest] -= (nDest+1) * dchi3    # ... ?
    deconv3 = np.zeros(nDest)
    rho = (degLPSD / 2) / degDS
    deconv3[1:nDest] = FormFunc3(chi3[1:nDest], rho)
    # Adjustment to avoid singularity of model function at x = 0
    # Note (deconv[0] + deconv[1] + ... + deconv[nDest-1]) * dx = 1)
    deconv3[0] = 1.0 / dchi3 - np.sum(deconv3[1:nDest])
    # Fourier Transform of Function To Be Deconvolved (Equatorial Model)
    ft_deconv3 = np.fft.rfft(deconv3)
    # Fourier Transform of Function To Be Convolved
    ft_conv3 = np.abs(ft_deconv3) + 0j

    #******************************************************************
    # Step 1-7: Fourier Transform about Artificial Shift Introduction
    # Function To Be Deconvolved is Dirac delta function "delta(x+k1)"
    # => (ft_inst3)
    #******************************************************************
    # Fourier transform of shift function to be deonvolved
    # Note: Shift by -k1 = +1.0 is introduced
    # It will cause (leave) Shift by -Kappa1 on 2Theta scale
    # Shift by -1 should be introduced to function to be deconvolved.
    # Shift by -k1 = +1.0 on Fourier transform is introduced.
    # Step interval of Fourier transform
    dxi3 = 1.0 / (dchi3 * nDest)
    xi3 = np.linspace(0, (nDest/2+1) * dxi3, int(nDest/2+1), endpoint=False)
    # xi[0] = 0, xi[1] = dxi, ..., 

    # ft_shift3 = np.exp(2 * np.pi * 1.0j * xi3) # Ida's erroneous (?) code
    ft_shift3 = np.exp(-2 * np.pi * 1.0j * xi3) # corrected by N. Sakurai

    # Fourier transform of instrumental function
    # ( to be used for error estimation )
    # print("shape of ft_shift3 : ",np.size(ft_shift3))
    # print("shape of ft_deconv3 : ",np.size(ft_deconv3))
    ft_inst3 = ft_deconv3 * ft_shift3/ ft_conv3
    # print("ft_shift3 = ",ft_shift3)
    # print("ft_inst3 = ",ft_inst3)

    #**********************************************************
    # Step 1-8: Here is the Core of DCT,
    # Deconvolution/Convolution Treatment on Fourier transform
    # (ft_eta3)/(ft_inst3) => (ft_zeta3)
    #**********************************************************
    # Deconvolution and Convolution Treatment
    ft_zeta3 = ft_eta3 / ft_inst3

    #*************************************
    # Step 1-9: Inverse Fourier Transform
    # (ft_zeta3) => (zeta3)
    #*************************************
    # Inverse Fourier transform of DCT data
    zeta3 = np.fft.irfft(ft_zeta3)

    #***********************************************
    # Step 1-10: Pick up Intensity from DCT data
    # on "chi3" scale by cubic spline interpolation
    # (chi3, zeta3) => (source_x, dest3_y)
    #***********************************************
    # Reset equidistant abscissa
    chi3 = np.linspace(chi0, chi0 + nDest * dchi3, nDest, endpoint = False)
    # Cubic spline interpolation, (chi,zeta) => (source_x, dest_y)
    f3 = interpolate.interp1d(chi3, zeta3, kind="cubic")
    dest3_y = f3(source_x)

    #**********************************************
    # Step 1-11: Rescale transform
    # (source_x, dest3_y) => (s_angle, d3_intensity)
    #**********************************************
    d3_intensity = dest3_y / source_corr
    dest = source
    dest[1] = d3_intensity

    #*******************************************************
    # Step 1-12 Fourier Treatment of Error Data, on Stage 1
    #*******************************************************
    dest = cmn.TreatErrors(error_type, source, source_x, index, source_corr, ft_inst3)

    #*******************************************************
    # Step 1-13 Pass "dest" as "source" for next stage ...
    #*******************************************************
    source = dest

    ##########################################
    # Stage 2: Correction of the 1st cumulant
    # (Shift Correction)
    ##########################################

    #******************************************************
    # Step 2-0: Divide Python List "source" to numpy.array
    # Attention: Missed in the former version 
    #******************************************************
    s_angle = np.array(source[0])
    s_intensity = np.array(source[1])
    if (error_type == '1'):
        s_error1 = np.array(source[2])
    elif (error_type == '2'):
        s_error2 = np.array(source[2])
    elif (error_type == '3'):
        s_error1 = np.array(source[2])
        s_error2 = np.array(source[3])
    d_intensity = np.array(s_intensity)
    
    #********************************************************
    # Step 2-1: Scale Transform for the 1st cumulant, "chi1"
    # (s_angle, s_intensity) => (source_x, source_y)
    #********************************************************
    # Abscissa values (Non-equidistantly separated)
    # kappa1 = kappas[1]    # 1st order cumulant of Delta2Theta in [rad.]
    k1 = -1    # 1st order cumulant of the model function on chi1-scale
    dchi1_d2Theta = k1 / kappa1
        # NOTE: the values should be similar to (k3 / kappa3)**(1/3) 
        # but should slightly be different.
    source_x = np.cumsum( dchi1_d2Theta )    # integration
    source_x *= (s_angle[1]-s_angle[0])
    # Ordinate correction factor
    source_corr = dchi1_d2Theta
    # Combine with geometrical correction
    source_corr = source_corr * cmn.g_corr(s_angle / 2)
    # Corrected intensity
    source_y = d3_intensity * source_corr

    #******************************************************************
    # Step 2-2: Preparation for Fourier treatment
    # Determine Marginal (Padding) Range, Total Points & Step Interval
    #******************************************************************
        # marginModel = np.sqrt(0.8*(1+1.25*rho**2))
        # margin: defines marginal (padding) range
        # margin = marginEquat * marginModel
        # marginEquat is the value read from the file'dct.cfg'
        # User can change the value in the current version.
        # NOTE: The use of the same value of margin may not be justified.
    index = cmn.indices(source_x, margin)
        # Derive total number and 1st,2nd,3rd anker points
        # dct_cmn.indices(x, w_margin) is defined in "dct_cmn.py"
        # x[..] = source_x :  abscissa (NumPy array)
    nSource = index[0] # number of source data points
    nValid = index[1] # index of 1st anker point on interpolated abscissa
    i2 = index[2] # index of 2nd anker point
    i3 = index[3] # index of 3rd anker point
    nDest = index[4] # total number of interpolated data
    nDest_2 = int(nDest/2) # half of the total number
    # Step Interval for "chi1" scale
    dchi1 = (source_x[nSource-1] - source_x[0]) / (nValid-1)
    # Prepare equidistant data for valid range
    chi0 = source_x[0]
    chi1 = np.linspace(chi0, chi0 + nValid * dchi1, nValid, endpoint = False)
    #*********************************************************
    # Step 2-3: 
    # Cubic spline interpolation of Data on Transformed Scale
    # (source_x, source_y) => (chi1, eta1)
    #*********************************************************
    # Cubic spline interpolation
    # eta = np.array(chi)
    f3 = interpolate.interp1d(source_x, source_y, kind="cubic")
    eta1 = f3(chi1) # (source_x, source_y) => (chi1, eta1)

    #********************************************
    # Step 2-4: 
    # Extrapolation for marginal (padding) range
    # (chi1, eta1) => (chi1, eta1)
    #********************************************
    # Fill marginal range with extrapolated values
    chi1 = np.linspace(chi0, chi0 + nDest * dchi1, nDest, endpoint = False)
    eta1 = np.resize(eta1, nDest)
    cmn.pad_margin(eta1, index)    # => "dct_common.py"

    #***********************************************
    # Step 2-5: Fourier Transform of Intensity Data
    # (eta1) => (ft_eta1)
    #***********************************************
    # Fourier transform of intensity
    ft_eta1 = np.fft.rfft(eta1)
    # Step interval of Fourier transform
    dxi1 = 1.0 / (dchi1 * nDest)
    # Values of abscissa
    xi1 = np.linspace(0, (nDest/2+1) * dxi1, int(nDest/2+1), endpoint=False)
    
    #***************************************************
    # Step 2-6: Fourier Transform of Shit function
    # Function To Be Deconvolved (Dirac delta function)
    # => (ft_inst1)
    #***************************************************
    # Fourier transform of function to be deonvolved
    # Note: Shift by k1 = -1 has been introduced on Stage 1
    # Shift by +1 should be introduced here.
    # Shift by k1 = -1.0 on Fourier transform is given by
    
    # ft_shift1 = np.exp(-2 * np.pi * 1.0j * xi1) # Ida's erroneous (?) code
    ft_shift1 = np.exp(2 * np.pi * 1.0j * xi1) # corrected by N. Sakurai
    
    ft_inst1 = ft_shift1

    #************************************************
    # Step 2-7: Shift Correction by Fourier method
    # Deconvolutional Treatment on Fourier transform
    # (ft_eta1)/(ft_inst1) => (ft_zeta1)
    #************************************************
    # ft_zeta1 = ft_eta1 / ft_inst1
    ft_zeta1 = ft_eta1 / ft_inst1
    
    #*************************************
    # Step 2-8: Inverse Fourier Transform
    # (ft_zeta1) => (zeta1)
    #*************************************
    zeta1 = np.fft.irfft(ft_zeta1)
    # Reset equidistant abscissa
    chi1 = np.linspace(chi0, chi0 + nDest * dchi1, nDest, endpoint = False)

    #******************************************************
    # Step 2-9: Pick up Intensity from shift treated data
    # on "chi3" scale by cubic spline interpolation
    # (chi1, zeta1) => (source_x, dest_y)
    #*****************************************************
    f3 = interpolate.interp1d(chi1, zeta1, kind="cubic")
    dest_y = f3(source_x)

    #***********************************************
    # Step 2-10: Rescale transform
    # (source_x, dest_y) => (2Theta, d_intensity)
    #***********************************************
    d_intensity = dest_y / source_corr
    dest = source
    dest[1] = d_intensity

    #*******************************************************
    # Step 2-11: Fourier Treatment of Error Data on Stage 2
    #*******************************************************
    dest = cmn.TreatErrors(error_type, source, source_x, index, source_corr, ft_inst1)

    return dest
    
