# -*- coding: utf-8 -*-
import numpy as np # import "numpy" module as "np"
import sys # import "sys" module to get command-line argument "argv"
import os # import "os" module to use "os.path.exists()"
import configparser # import "configparser"

import dct_common as cmn  # import 'dct_common.py'
import xray  # import 'xray.py' to treat spectroscopic profile of source X-ray
import axial  # import 'axial.py' to treat axial-divergence aberration
import equatorial  # import 'equatorial.py' to treat eqatorial aberration
import trnspr  # import 'trnspr.py' to treat sample-transparency aberration
import cntmn as cntmn  # import 'cntmn.py' for decontamination

#################
# 1. PREPARATION
#################

###################################
# 1-1 Treat Command-Line Arguments
###################################
argvs = sys.argv # list of command-line arguments
argc = len(argvs) # number of arguments + 1
# DPATH = os.path.dirname(sys.argv[0]) # directory of the EXE file
DPATH = ''
if (argc == 1):	# in case no arguments
	argvs = argvs + [os.path.join(DPATH, '00raw.csv')]
	argc = 2
if (argc == 2): # in case 1 argument
	argvs = argvs + [os.path.join(DPATH, '01dct.csv')]
	argc = 3
# In case 'filename.txt' is used ...
# f = open('filename.txt')
# argvs = [s.strip() for s in f.readlines()]
# f.close()
# originally command-line arguments
# argc = len(argvs) # number of arguments
	# argvs[1] is input file name
	# argvs[2] is output file name

###################################################
# 1-2 Read Configuration File "dct.cfg" 
###################################################
# Read dct.cfg
dct_config = configparser.ConfigParser()
CFG_FILE = os.path.join(DPATH, 'dct.cfg')
if os.path.exists(CFG_FILE):
	dct_config.read(CFG_FILE, encoding = 'utf-8')
else:
	sys.stderr.write(CFG_FILE + " is not found.\n")
	sys.exit(2)

# Which Way to Treat Error
error_type = dct_config.get('Control','error_type')
# Line Numbers of Header Part in Data File
headerLines = dct_config.get('Control','headerLines')
# Skip Spectroscopic Treatment
skipXray = '0'
if (dct_config.has_option('Control','skipXray')):
	skipXray = dct_config.get('Control','skipXray')
# Skip Axial Treatment
skipAxial = '0'
if (dct_config.has_option('Control','skipAxial')):
	skipAxial = dct_config.get('Control','skipAxial')
# Skip Equatorial Treatment
skipEquatorial = '0'
if (dct_config.has_option('Control','skipEquatorial')):
	skipEquatorial = dct_config.get('Control','skipEquatorial')
# Skip Transparency Treatment
skipTransparency = '0'
if (dct_config.has_option('Control','skipTransparency')):
	skipTransparency = dct_config.get('Control','skipTransparency')
# Skip Contamination Treatment
skipContamination = '0'
if (dct_config.has_option('Control','skipContamination')):
	skipContamination = dct_config.get('Control','skipContamination')

# Horizontal Error (deg.)
hrz_error = dct_config.getfloat('Control','hrz_error')
# Axial-Divergence Angle (deg.)
degAxial = dct_config.getfloat('Control','degAxial')
# Divergence Slit Open Angle (deg.)
degDS = dct_config.getfloat('Control','degDS')
# View Angle of Si-strip X-ray Detector (deg.)
degLPSD = dct_config.getfloat('Control','degLPSD')
# Ratio of Sample Width to Gonoimeter Radius
specW_R = dct_config.getfloat('Control','specW_R')
# Ratio of Penetration-Depth to Goniometer Radius
muR = dct_config.getfloat('Control','muR')

###################################################
# 1-3 Read angle, intensity [ & error ] data file
###################################################
# 1st argument argvs[1] should be data file name
if not os.path.exists(argvs[1]):
	sys.stderr.write(argvs[1] + " is not found.\n")
	sys.exit(3)
# Read 1st line of data file
f = open(argvs[1], 'r')
line = f.readline()
nSkip = int(headerLines)
# What is 1st line of data file ?
#if (line == '[Measurement conditions]\n'):
	# If it is [Measurement conditions] 
	# (PANalytical Data Collector format),
	# skip 30 lines
#	nSkip = 30
#else:
	# If the 1st line is not "PANalytical Data Collector format",
	# do not skip lines
#	nSkip = 0

# Read data file
source_data = np.loadtxt(argvs[1],dtype='float',delimiter=',',skiprows=nSkip)

# Extract angle, intensity (, error) data
source_angle = source_data[:,0]	# 0th column is angle
source_intensity = source_data[:,1]	# 1st column is intensity
nSource = len(source_angle)
# If number of columns are equal to or larger than 3,
if (source_data.shape[1] >= 3):
	source_error = source_data[:,2]	# 2nd column is error
else:
	# Error estimation
	angular_error = hrz_error # estimated angular error (º)
	source_error = np.array(source_intensity) # reserve array
	dest_intensity = np.array(source_intensity) # reserve array
	dest_error = np.array(source_intensity) # reserve array
	dx = source_angle[1] - source_angle[0]
	y0 = source_intensity[0]
	y1 = source_intensity[1]
	err = (y1 - y0) / dx * angular_error
	source_error[0] = np.sqrt(y0 + np.square(err))
	for index in range(1,nSource-1):
	    dx = source_angle[index+1] - source_angle[index-1]
	    y0 = source_intensity[index-1]
	    y1 = source_intensity[index]
	    y2 = source_intensity[index+1]
	    err = (y2 - y0) / dx * angular_error
	    source_error[index] = np.sqrt(y1 + np.square(err))
	dx = source_angle[nSource-1] - source_angle[nSource-2]
	y0 = source_intensity[nSource-2]
	y1 = source_intensity[nSource-1]
	err = (y1 - y0) / dx * angular_error
	source_error[nSource-1] = np.sqrt(y1 + np.square(err))

# Save Copy of Source Data
source_intensity0 = source_intensity.copy() # make a copy
source_error0 = source_error.copy() # make a copy

##############
# Source Data
##############
if (error_type == '0'):
	source = [source_angle, source_intensity]
if ((error_type == '1') or (error_type == '2')):
	source = [source_angle, source_intensity, source_error]
if (error_type == '3'):
	source = [source_angle, source_intensity, source_error, source_error]
# Make a copy
source0 = source

##########################################
# 2. TREAT SPECTROSCOPIC PROFILE OF X-RAY
##########################################
if (skipXray != '1'):
	#*****************************
	# Treat Spectroscopic Profile
	#*****************************
	# Deconvolution-Convolution Treatment about Spectroscopic Profile
	dest = xray.treat(source, DPATH)
	dest_intensity = np.array(dest[1]) # output intensity
else:
	dest = source
	
#######################################
# 3. TREAT AXIAL-DIVERGENCE ABERRATION 
#######################################
if ((skipAxial != '1') and (degAxial > 0.0)):
	#***********************************************
	# 3-1 Treat Lower-Angle Part of Axial-Divergnce 
	#***********************************************
	source = dest	# Copy "dest" to "source" 
	iAxial = -1	# iAxial = -1: lower-angle part 
	dest = axial.treat(iAxial, source, DPATH)
	#*************************************************
	# 3-2 Treat Higher-Angle Part of Axial-Divergence 
	#*************************************************
	source = dest	# Copy "dest" to "source" 
	iAxial = 1	# iAxial = 1: higher-angle part
	dest = axial.treat(iAxial, source, DPATH)
else:
	# Copy "source" to "dest"
	dest = source

####################################
# 4. TREAT EQUATORIAL ABERRATION 
####################################
if ((skipEquatorial != '1') and (degDS > 0.0)):
	#********************************
	# Treat Equatorial Aberration 
	#********************************
	# Copy "dest" to "source"
	source = dest
	# Treat Equatorial Aberration
	dest = equatorial.treat(source, DPATH) 
else:
	# Copy "source" to "dest"
	dest = source

##########################################
# 5. TREAT SAMPLE-TRANSPARENCY ABERRATION 
##########################################
if ((skipTransparency != '1') and (muR > 0.0)):
	#**************************************
	# Treat Sample-Transparency Aberration 
	#**************************************
	# Copy "dest" to "source"
	source = dest
	dest = trnspr.treat(source, DPATH) 
else:
	dest = source

#####################
# 6. DECONTAMINATION 
#####################
if (skipContamination != '1'):
	#*****************************************
	# Treat Sub-peaks Caused by Contamination 
	#*****************************************
	# Copy "dest" to "source"
	source = dest
	dest = cntmn.treat(source, DPATH) 
else:
	dest = source

################################
# Save Data to Text (.csv) File
################################
save = np.array(dest).T # transpose matrix
if (error_type == '0'):
	np.savetxt(argvs[2], save, delimiter=",",fmt=["%.3f","%.2f"])
elif ((error_type == '1') or (error_type == '2')):
	np.savetxt(argvs[2], save, delimiter=",",fmt=["%.3f","%.2f,%.2f"])
elif (error_type == '3'):
	np.savetxt(argvs[2], save, delimiter=",",fmt=["%.3f","%.2f,%.2f,%.2f"])
