# -*- coding: utf-8 -*-
##########################################
# trnspr.py 
# to treat Sample-Transparency Aberration
# originally coded by T. Ida
# corrected by T. Ida, Feb. 3, 2021
# corrected by T. Ida, Jun. 22, 2021
# modified by T. Ida, Sep. 8, 2021
##########################################
import numpy as np # import "numpy" modeule as "np"
from scipy import interpolate # import "interpolate" from "scipy"
from scipy import special # import "special" from "scipy"
import os.path # import "os.path" module
import configparser # import "configparser"
import dct_common as cmn # import "dct_common.py" as "cmn"

########################################################
# calc_kappas()
# Calculate Cumulants of Sample-Transparency Aberration
########################################################
def calc_kappas(degTwot, muR, t_R):
    # degTwot: 2Theta in [deg.] (NumPy array)
    # muR = mu * R, (float number)
    #    mu is linear attenuation coefficient, 
    #    R is the goniometer radius
    # t_R = t / R, (float number)
    #    t is the thickness of the sample
    #    R is the goniometer radius
    theta = degTwot * np.pi / 360    # (NumPy array)
    u = 2 * t_R * np.cos(theta)    # (NumPy array)
    g = np.sin(2 * theta) / (2 * muR)    # (NumPy array)
    exp_u_g = np.exp(-u/g)    # (NumPy array)
    # Calculate Raw (Non-Central) Moments up to 4th order
    # Attention: erroneous in the former version
    s0 = 1 - exp_u_g
    s1 = -(-u)*exp_u_g - g*s0
    s2 = -(-u)**2*exp_u_g - g*2*s1
    s3 = -(-u)**3*exp_u_g - g*3*s2
    s4 = -(-u)**4*exp_u_g - g*4*s3
    m1 = s1 / s0
    m2 = s2 / s0
    m3 = s3 / s0
    m4 = s4 / s0
    # Calculate Cumulants up to 4th order
    kappa1 = m1
    kappa2 = m2 - m1**2
    kappa3 = m3 - 3*m2*m1 + 2*m1**3
    kappa4 = m4 - 4*m3*m1 - 3*m2**2 + 12*m2*m1**2 - 6*m1**4
    # Convert to Reduced Cumulants in [deg.]
    kappa1 = 180.0/np.pi*kappa1
    kappa2 = 180.0/np.pi*np.sqrt(kappa2)
    kappa3 = 180.0/np.pi*np.sign(kappa3)*np.abs(kappa3)**(1/3)
    kappa4 = 180.0/np.pi*np.sign(kappa4)*np.abs(kappa4)**(1/4)
    kappas = np.array([s0, kappa1,kappa2,kappa3,kappa4])
    return([s0,kappa1,kappa2,kappa3,kappa4])
    
############################################################
# Definition of Fourier transform of deconvolution function
# As the instrumental function is "exp(chi) [chi < 0]",
# Fourier transform should be
#  int_{-inf}^{0} exp(chi) * exp(2πi xi chi) dchi
#  = int_{-inf}^{0} exp((1 + 2πi xi) chi) dchi
#  = [exp((1 + 2πi xi) chi) / (1 + 2πi xi) ]_{k=-inf}^{k=0}
#  = 1 / (1 + 2πi xi)
# NOTE: depends on definition of the Fourier transform
############################################################
def fFtDeconv(xi):
    ans = 1.0 / (1.0 - 2 * np.pi * 1j * xi)
    return ans

################################################
# "trnspr.treat"
# TREATMENT OF SAMPLE-TRANSPARENCY, MAIN BODY
################################################
def treat(source, DPATH):
    # source: Source Data (Python List)
    #    [s_angle, s_intensity],
    #    [s_angle, s_intensity, s_error1], or
    #    [s_angle, s_intensity, s_error1, s_error2]
    # DPATH: Path of Configuration File

    #***********************************
    # Read Configuration File "dct.cfg"
    #***********************************
    CFG_FILE = os.path.join(DPATH, 'dct.cfg')
    dct_config = configparser.ConfigParser()
    dct_config.read(CFG_FILE, encoding = 'utf-8')
    error_type = dct_config.get('Control','error_type')
    # muR = (linear attenuation coefficient)*(goniometer radius)
    muR = dct_config.getfloat('Control','muR')
    # t_R = (sample thicknes)/(goniometer radius)
    t_R = dct_config.getfloat('Control','t_R')
    # Margin for Fourier-based Deconvolutional Treatment
    # Attention: not utilized in some previous versions.
    marginTrnspr = dct_config.getfloat('Control','marginTrnspr')
    
    ##############################################################
    # Stage 1: Compensation of the truncation by finite thickness 
    # (Recovers the intensity lost by truncation)
    ##############################################################

    #******************************************************
    # Step 1-0: Divide Python List "source" to numpy.array
    #******************************************************
    s_angle = np.array(source[0])
    s_intensity = np.array(source[1])
    if (error_type == '1'):
        s_error1 = np.array(source[2])
    elif (error_type == '2'):
        s_error2 = np.array(source[2])
    elif (error_type == '3'):
        s_error1 = np.array(source[2])
        s_error2 = np.array(source[3])
    d_intensity = np.array(s_intensity)

    #-----------------------------------------------
    # Optional Output for 'SkipTransparency' == '2'
    #-----------------------------------------------
    skipTransparency = '0'
    if (dct_config.has_option('Control','skipTransparency')):
        skipTransparency = dct_config.get('Control','skipTransparency')
    if (skipTransparency == '2'):
        kappas = calc_kappas(s_angle, muR, t_R)
        CUMULANT_FILE = 'cumulants_transparency.csv'
        save_cumulant = np.insert(kappas, 0, s_angle, axis=0)
        save_cumulant = np.array(save_cumulant).T    # transpose matrix
        fmt_list = ["%.3f","%.5e","%.5e","%.5e","%.5e","%.5e"]
        np.savetxt(CUMULANT_FILE, save_cumulant, delimiter=",",fmt=fmt_list)
    
    #****************************************************
    # Step 1-1: Scale Transform for the finite thickness
    # s_angle (2Theta in deg.) => chiS
    # s_intensity => etaS0
    #****************************************************
    vEPS = 1.0E-6
    nPoint = s_angle.size
    
    theta = s_angle * np.pi / 360
    chiS = 1.0/t_R*np.log(2/(1-np.tan(theta/2))-1)
    u = 2*t_R*np.cos(theta)
    g = np.sin(2*theta) / (2 * muR)
    nChi = np.ceil(np.log(g[nPoint-1]/u[nPoint-1]/vEPS))
    nChi = nChi.astype(np.int64)
    # Ordinate Corrections
    # (Lorentz, polarization & powder diffractometry corrections)
    source_corr = u * cmn.g_corr(s_angle / 2)
    # Corrected & Transdformed Intensity
    # Attention: Intensity correction factor 
    # defined in "dct_common.py" should be multiplied 
    # to correct the intensity,
    # but divided in some previous versions...
    etaS0 = s_intensity * source_corr

    #*********************************************
    # Step 1-2: Compensation for Finite thickness
    #*********************************************
    # for iChi in range(nChi): # for iChi = 0 to nChi-1
    #    chiS_i = chiS - iChi - 1 # shifted by -1, -2, -3, ...
    #    etaS_i = etaS0 * exp(-u/g * (iChi+1) ) # attenuate 
    #    f1 = interpolate.interp1d(chiS_i, etaS_i, kind="linear")
    #    etaS1 += f1(chiS)
    iChi = np.arange(nChi).repeat(nPoint).reshape(nChi,nPoint)
    chiS_i = np.tile(chiS,nChi).reshape(nChi,nPoint)
    chiS_i -= iChi
    chiS_i[:,nPoint-1] = chiS[nPoint-1]
    etaS_i = np.tile(etaS0,nChi).reshape(nChi,nPoint)
    exp_i = np.exp(-iChi*(np.tile(u/g, nChi).reshape(nChi,nPoint)))
    etaS_i = etaS_i * exp_i
    etaS1_i = etaS_i
    for i in range(nChi):
        f1 = interpolate.interp1d(chiS_i[i], etaS_i[i])
        etaS1_i[i] = f1(chiS)
    etaS1 = np.sum(etaS1_i, axis=0)

    #**********************************************
    # Step 1-3: Rescale transform
    # (chiS, etaS1) => (s_angle, d_intensity)
    #**********************************************
    d_intensity = etaS1 / source_corr
    dest = source
    dest[1] = d_intensity

    #*******************************************************
    # Step 1-4 Fourier Treatment of Error Data, on Stage 1
    # (not used in the current version)
    #*******************************************************
    # dest = cmn.TreatErrors(error_type, source, source_x, index, source_corr, ft_inst)
    # dest = cmn.TreatErrors(error_type, source, source_x, index, source_corr, ft_inst)

    #*******************************************************
    # Step 1-5 Pass "dest" as "source" for the 2nd stage ...
    #*******************************************************
    source = dest

    ##########################################
    # Stage 2: Correction of the Extinction
    # (Symmetrization about decay effect)
    ##########################################
    
    #******************************************************
    # Step 2-0: Divide Python List "source" to numpy.array
    # Attention: Missed in some previous versions
    #******************************************************
    s_angle = np.array(source[0])
    s_intensity = np.array(source[1])
    if (error_type == '1'):
        s_error1 = np.array(source[2])
    elif (error_type == '2'):
        s_error2 = np.array(source[2])
    elif (error_type == '3'):
        s_error1 = np.array(source[2])
        s_error2 = np.array(source[3])
    d_intensity = np.array(s_intensity)

    #********************************************************
    # Step 2-1: Scale Transform for the decay, "chiT"
    # (s_angle, s_intensity) => (source_x, source_y)
    #********************************************************
    # Abscissa values (Non-equidistantly separated) (chiT)
    source_x = 2 * muR *np.log(np.tan(theta))
    # Ordinate correction factor
    source_corr = g * cmn.g_corr(s_angle / 2)
    # Corrected intensity (etaT)
    source_y = s_intensity * source_corr

    #******************************************************************
    # Step 2-2: Preparation for Fourier treatment
    # Determine Marginal (Padding) Range, Total Points & Step Interval
    #******************************************************************
    marginModel = 1
        # margin: defines marginal (padding) range
    margin = marginTrnspr * marginModel
        # marginTrnspr is the value read from the file'dct.cfg'
        # User can change the value in the current version.
        # NOTE: The use of the same value of margin may not be justified.
    index = cmn.indices(source_x, margin)
        # Derive total number and 1st,2nd,3rd anker points
        # dct_cmn.indices(x, w_margin) is defined in "dct_cmn.py"
        # x[..] = source_x :  abscissa (NumPy array)
    nSource = index[0] # number of source data points
    i1 = index[1] # index of 1st anker point on interpolated abscissa
    i2 = index[2] # index of 2nd anker point
    i3 = index[3] # index of 3rd anker point
    nDest = index[4] # total number of interpolated data
    nDest_2 = int(nDest/2) # half of the total number
    # Step Interval on "chiT" scale
    dx = (source_x[nSource-1] - source_x[0]) / i1
    # Prepare equidistant data for valid range
    nValid = int((source_x[nSource-1] - source_x[0]) / dx)
    chi0 = source_x[0]
    chi = np.linspace(chi0, chi0 + nValid * dx, nValid, endpoint = False)

    #*********************************************************
    # Step 2-3: 
    # Cubic spline interpolation of Data on Transformed Scale
    # (source_x, source_y) => (chi, eta)
    #*********************************************************
    # Cubic spline interpolation
    # eta = np.array(chi)
    f3 = interpolate.interp1d(source_x, source_y, kind="cubic")
    eta = f3(chi) # (source_x, source_y) => (chi, eta)

    #********************************************
    # Step 2-4: 
    # Extrapolation for marginal (padding) range
    # (chi, eta) => (chi, eta)
    #********************************************
    # Fill marginal range with extrapolated values
    chi = np.linspace(chi0, chi0 + nDest * dx, nDest, endpoint = False)
    eta = np.resize(eta, nDest)
    iLin = np.linspace(0, i3-i2, i3-i2)
    etaL = source_y[nSource - 1]
    etaR = source_y[0]
    eta[i1:i2] = etaL
    eta[i2:i3] = etaL + (etaR - etaL) * iLin[0:i3-i2] / (i3-i2)
    eta[i3:nDest] = etaR

    #***********************************************
    # Step 2-5: Fourier Transform of Intensity Data
    # (eta) => (ft_eta)
    #***********************************************
    # Fourier transform of intensity
    ft_eta = np.fft.fft(eta)

    #***************************************************
    # Step 2-6: Fourier Transform of Decay function
    # => (ft_inst)
    #***************************************************
    # Fourier transform of function to be deonvolved
    inst = np.linspace(-nDest*dx, 0, nDest)
    inst = np.exp(inst)
    ft_inst = np.fft.fft(inst)
    # Fourier transform of function to be convolved
    conv = np.zeros(nDest)
    # Set start point to be zero・・
    ft_inst = ft_inst / np.abs(ft_inst)

    #************************************************
    # Step 2-7: Decay Correction by Fourier method
    # Deconvolutional Treatment on Fourier transform
    # (ft_eta)/(ft_inst) => (ft_eta)
    #************************************************
    ft_eta = ft_eta / ft_inst

    #*************************************
    # Step 2-8: Inverse Fourier Transform
    # (ft_eta) => (zeta)
    #*************************************
    zeta = np.real(np.fft.ifft(ft_eta))
    # Reset equidistant abscissa
    chi = np.linspace(chi0, chi0 + nDest * dx, nDest, endpoint = False)

    #******************************************************
    # Step 2-9: Pick up Intensity from shift treated data
    # on "chi3" scale by cubic spline interpolation
    # (chi, zeta) => (source_x, dest_y)
    #*****************************************************
    fT = interpolate.interp1d(chi, zeta, kind="cubic")
    dest_y = fT(source_x)

    #***********************************************
    # Step 2-10: Rescale transform
    # (source_x, dest_y) => (2Theta, d_intensity)
    #***********************************************
    d_intensity = dest_y / source_corr
    dest = source
    dest[1] = d_intensity

    #*******************************************************
    # Step 2-11: Fourier Treatment of Error Data on Stage 2
    #*******************************************************
    dest = cmn.TreatErrors(error_type, source, source_x, index, source_corr, ft_inst)

    return dest # Return Python list
    
